classdef LJ < handle
    
    % LabJack class for U3, UE9, or T7
    %
    % Under construction...
    %
    % NOTE: To access streaming data on the fly, at the frequency 
    % 1/chunk_time_s, you can add a listener for the 'DataAvailable' event, 
    % which is broadcasted every time a chunk of data has been collected, 
    % e.g.:
    %   L = LJ;
    %   addlistener(L,'DataAvailable',...
    %   @(src, event)update_lines(src, event, additional_var_1, ..., additional_var_n));
    % where src gives access to L from within the callback.
    %
    % TODO: stop stream, if running, before clearing object
    % TODO: make the whole thing compatible with the data acquisition toolbox syntax and methodology
    %
    % Author: DJ van Gerwen
    % Created: 2014-10-10
    
    
    % NOTE: Starting from Jan 2016 (or LJclass_v60.m), this file is under
    % Git source control.
    % Useful commands (from windows command window (cmd) in current dir, or
    % in the matlab cmd using the bash operator (e.g. !git init) ):
    %   git help
    %   git init
    %   git status
    %   git branch
    %   git checkout
    %   git add some_filename.m
    %   git commit -m "some description here"
    % Check the reference: [chacon 2015 - Pro GIT] available on the web for free.
    % Although matlab supports git, creating a repository is much easier
    % from the windows cmd. Just go into the directory (in cmd), call git
    % init, then add the files you want using git add, and make the first
    % commit using git commit. Also branching is a bit more straightforward
    % using  the branch and checkout commands.
    
    %% Properties
    properties (Hidden=true)
        UDassembly
        UDobject
        Massembly
        Mconstants
        lj_handle
        timer_handle
        backlog_string
        last_error
        stream_rate_achieved_Hz
        chunk_data
        chunk_size
    end
    
    
    properties (Access=public, SetObservable)
        device_type = '' % Options: see supported_device_types property below
        device_id = 'ANY' % Options: 'ANY', serial number, IP-address, or device name
        connection_type = 'USB' % Options: 'ANY', 'USB', 'TCP', 'ETHERNET', 'WIFI' (depending on device_type)
        AIN_list = 0 % List of channel IDs [-]
        AIN_range_V % Set ranges for individual channels (for T7: +/-0.1V, +/-1V, or +/-10V)
        % TODO: auto-adjust the AIN_range_V list when AIN_list changes length (maybe use set method or listener)
    end
    properties (Dependent=true, SetAccess=private)
        AIN_state_V % Value of specified AIN ports (NOTE: Labjack is contacted whenever this property is requested)
    end
    properties (Access=public, SetObservable)
        DOUT_list % Digital output indices (FIO ports) [-]
        DOUT_state % Digital output values [0/1]
        DIN_list % Digital input indices (FIO ports) [-]
    end
    properties (Dependent=true, SetAccess=private)
        DIN_state % Digital input values [0/1]
    end
    properties (Access=public, SetObservable)
        DAC_list % Analog output indices [-]
        DAC_state_V % Analog output values [V]
        resolution_bits = 12; % [bits] AIN resolution (fixed for U3?)
        relay_list % Listof FIO ports to use for relay switching (sinking configuration)
        relay_state %  [0/1] Switch external relay onnected to Vs (relay+) and FIOi (relay-) (sinking configuration)
        PWM_list % List of FIO ports to use for PWM (T7 can use FIO 0, 2, and 3, but NOT FIO 1)
        PWM_state % List of normalized PWM duty cycles (0 to 1 corresponds with minimum to maximum of range)
        PWM_duty_cycle_limits % If empty, maximum range is used
        PWM_carrier_frequency_Hz = 45 % Desired PWM carrier frequency
        scan_rate_requested_Hz = 200 % [Hz] Number of times each individual channel is sampled per second, so the total sample rate is labjack.scan_rate_requested_Hz*labjack.number_of_channels and must be smaller than 50kHz for 12bit resolution
    end
    properties (SetAccess=private)
        scan_rate_achieved_Hz
        backlog_device
        backlog_driver
    end
    properties (Access=public, SetObservable)
        chunk_time_s = 1 % [s] Wait-time for retrieving data [s] NOTE: Chunks need to be sufficiently large if high sample rates are required
        buffer_time_s = 5 % [s] Stream buffer size (driver) NOT USED FOR T7!!!
        stream_data_V
        stream_data_maxlength = 1e7 % FIFO truncation occurs when stream history exceeds this size
        show_backlogs = 0
        
    end
    
    properties (Dependent=true, SetAccess=private, Hidden=true)
        number_of_AIN
        number_of_stream_channels
        AIN_names
        stream_rate_requested_Hz % Number of times the entire list of channels is sampled per second
        expected_scans_per_chunk
        expected_samples_per_chunk
        buffer_size
        PWM_duty_cycle % Only used for T7
    end
    
    properties (SetAccess=private, Hidden=true)
        stream_started = 0
        supported_device_types = {'T7', 'UE9', 'U3'}
        PWM_clock_source %= 0 % Labjack T7 clock0 is 32bit, so counts from 0 to roll_value (maximum 2^32-1) at 80MHz
        PWM_clock_frequency_Hz %= 80e6 % Labjack T7 clock0 frequency is 80MHz
        PWM_clock_divisor %= 1 % A divisor of 1 allows frequencies from 80e6/(2^32-1)=0.019Hz to 80MHz
        PWM_clock_roll_value % Only used for T7 % The clock counter is reset when it reaches the roll value (0<=roll_value<=2^32-1 for 32bit clock)
    end
    
    events
        DataAvailable % Similar to the daq-toolbox event
    end
    
    %% Methods
    
    %     methods (Access = protected)
    %         function propgrp = getPropertyGroups(~)
    %             proplist = {'device_type','AIN_list','FIO_list','FIO_state','DAC_list','DAC_state_V','scan_rate_requested_Hz','scan_rate_achieved_Hz','chunk_time_s','buffer_time_s','buffer_size','stream_data_V'};
    %             propgrp = matlab.mixin.util.PropertyGroup(proplist);
    %         end
    %     end
    
    
    methods (Access = public, Hidden = true)
        
        % CONSTRUCTOR
        function L = LJ(device_type, device_id, connection_type, AIN_list, ...
                scan_rate_requested_Hz, chunk_time_s, buffer_time_s)
            
            % TODO: check why device_id is not updated for UE9...
            % TODO: check why get.AIN_state_V  works with UE9 over USB, but gives
            % error when using ETHERNET mode... (error message: No method 'eGet' with matching signature found for class 'LabJack.LabJackUD.LJUD'.)
            
            % Device selection
            if exist('device_type','var')
                if ~isempty(device_type)
                    L.device_type = device_type;
                end
            end
            if exist('device_id','var')
                if ~isempty(device_id)
                    L.device_id = device_id;
                end
            end
            if exist('connection_type','var')
                if ~isempty(connection_type)
                    L.connection_type = connection_type;
                end
            end
            
            % Independent LabJack streaming parameters
            if exist('AIN_list','var')
                if ~isempty(AIN_list)
                    L.AIN_list = AIN_list;
                end
            end
            if exist('scan_rate_requested_Hz','var')
                if ~isempty(scan_rate_requested_Hz)
                    L.scan_rate_requested_Hz = scan_rate_requested_Hz;
                end
            end
            if exist('chunk_time_s','var')
                if ~isempty(chunk_time_s)
                    L.chunk_time_s = chunk_time_s;
                end
            end
            if exist('buffer_time_s','var')
                if ~isempty(buffer_time_s)
                    L.buffer_time_s = buffer_time_s;
                end
            end
            
            % Initialize labjack
            try
                L.open_labjack()
            catch err
                disp(err)
                error('Cannot open labjack. Check labjack type (e.g. try object = LJ(''UE9'') ).')
            end
            L.configure_labjack()
            try
                L.PWM_carrier_frequency_Hz = L.PWM_carrier_frequency_Hz; % Updates the labjack PWM settigns
            catch err
                disp(err.message)
                disp('LJ object not fully functional.')
            end
            
        end % Constructor
        
    end
    
    methods % SET-METHODS
        %#ok<*MCSUP>
        % NOTE: property initialization order dependency (upon loading a
        % saved object) is not an issue, because states must be reset
        % to zero after loading the object anyway.
        % NOTE: could do this in a listener callback instead
        % NOTE: Any change to a list should result in the corresponding states being reset
        function set.PWM_list(L, new_PWM_list)
            % TODO: prevent duplicates in FIO_list, relay_list, and PWM_list
            % First disable all PWM ports in the current list
            switch L.device_type
                case 'T7'
                    for i = 1:length(L.PWM_list)
                        % Could use eWriteNames() instead of the loop, but looping is just as easy
                        name_i = sprintf('DIO%u_EF_', L.PWM_list(i));
                        LabJack.LJM.eWriteName(L.lj_handle, [name_i 'ENABLE'], 0); % Disable any extended functionality running on this pin
                    end
                    
                    % Impose 0,2,3 range (T7 has 4 FIO screw-terminals, for some reason FIO1 cannot be used... see this table: https://labjack.com/support/datasheets/t7/digital-io/extended-features)
                    L.PWM_list = intersect(new_PWM_list, [0 2 3]);
                    
                case 'UE9'
                    
                    % Impose FIO range (TODO: for the UE9, PWM_list must
                    % start at 0 and must be consecutive)
                    L.PWM_list = intersect(new_PWM_list, 0:3);
                    
            end
            L.PWM_state = zeros(size(L.PWM_list)); % Set all PWM duty cycles to zero
            switch L.device_type
                case 'T7'
                    for i = 1:length(L.PWM_list) % NOTE: if list is empty, ports remain disabled
                        % Could use eWriteNames() instead of the loop, but looping is just as easy
                        name_i = sprintf('DIO%u_EF_',L.PWM_list(i));
                        LabJack.LJM.eWriteName(L.lj_handle, [name_i 'ENABLE'], 0); % Disable any extended functionality running on this pin (in case new pin is selected)
                        LabJack.LJM.eWriteName(L.lj_handle, [name_i 'INDEX'], 0); % Configure pin for PWM
                        LabJack.LJM.eWriteName(L.lj_handle, [name_i 'OPTIONS'], L.PWM_clock_source); % Select clock source
                        LabJack.LJM.eWriteName(L.lj_handle, [name_i 'CONFIG_A'], 0); % Init duty cycle to zero
                        LabJack.LJM.eWriteName(L.lj_handle, [name_i 'ENABLE'], 1); % Start PWM
                        fprintf(1,'PWM signal enabled on DIO%u\n',i);
                    end
                case 'UE9'
                    % Call update_PWM_carrier_frequency by re-setting the
                    % property value (Quick and Dirty solution)
                    L.PWM_carrier_frequency_Hz = L.PWM_carrier_frequency_Hz;
            end
            
        end
        
        function set.PWM_state(L, new_PWM_state)
            if length(new_PWM_state) == length(L.PWM_list)
                L.PWM_state = new_PWM_state;
                L.update_PWM_duty_cycle;
            else
                error('Number of states must match number of ports in the list.')
            end
        end
        function set.PWM_carrier_frequency_Hz(L, new_carrier_frequency_Hz)
            L.PWM_carrier_frequency_Hz = ...
                L.update_PWM_carrier_frequency(new_carrier_frequency_Hz);
        end
        function set.relay_list(L,new_relay_list)
            % TODO: prevent duplicates in FIO_list and relay_list and PWM_list
            L.relay_state = zeros(size(L.relay_list)); % Reset existing relay states to zero
            L.relay_list = intersect(new_relay_list, 0:3); % Impose 0-3 range (T7 has 4 FIO screw-terminals)
            L.relay_state = zeros(size(L.relay_list)); % Set new relay states to zero
        end
        function set.relay_state(L,new_relay_state)
            if length(new_relay_state) == length(L.relay_list)
                L.relay_state = new_relay_state;
                L.toggle_relay;
            else
                error('Number of states must match number of ports in the list.')
            end
        end
        function set.DAC_list(L, new_DAC_list)
            L.DAC_state_V = zeros(size(L.DAC_list)); % Reset existing DAC states to zero
            L.DAC_list = intersect(new_DAC_list, 0:1); % Impose 0-1 range (U3HV and T7 only have 2 DAC ports)
            L.DAC_state_V = zeros(size(L.DAC_list)); % Set new DAC states to zero
        end
        function set.DAC_state_V(L, new_DAC_state_V)
            if length(new_DAC_state_V) == length(L.DAC_list)
                L.DAC_state_V = min(max(new_DAC_state_V, 0), 5); % Impose 0V-5V range (U3HV)
                L.set_aout;
            else
                error('Number of states must match number of ports in the list.')
            end
        end
        function set.DIN_list(L, new_DIN_list)
            % TODO: prevent duplicates in DOUT_list, DIN_list, and relay_list
            
            % Limit to available channels
            switch L.device_type
                case 'U3'
                    L.DIN_list = intersect(new_DIN_list, 4:7); % Impose 4-7 range (U3HV)
                case 'UE9'
                    L.DIN_list = intersect(new_DIN_list, 0:7); % Impose 0-7 range (UE9 NOTE: only 0:3 are screw terminals)
                case 'T7'
                    L.DIN_list = intersect(new_DIN_list, 0:22); % Impose 0-22 range (T7)
            end
            L.DIN_state; % Initialize state vector
            L.configure_labjack; % Reconfigure to include digital channels to stream
        end
        function set.DOUT_list(L, new_DOUT_list)
            % TODO: prevent duplicates in DOUT_list, DIN_list, and relay_list
            
            % Set state to zero for all FIO in current list (in case
            % channels are removed from list)
            L.DOUT_state = false(size(L.DOUT_list));
            
            % Limit to available channels
            switch L.device_type
                case 'U3'
                    L.DOUT_list = intersect(new_DOUT_list, 4:7); % Impose 4-7 range (U3HV)
                case 'UE9'
                    L.DOUT_list = intersect(new_DOUT_list, 0:7); % Impose 0-7 range (UE9 NOTE: only 0:3 are screw terminals)
                case 'T7'
                    L.DOUT_list = intersect(new_DOUT_list, 0:22); % Impose 0-22 range (T7)
            end
            
            % Set state to zero for all FIO in new list
            L.DOUT_state = false(size(L.DOUT_list));
            
        end
        function set.DOUT_state(L, new_DOUT_state)
            if length(new_DOUT_state) == length(L.DOUT_list)
                L.DOUT_state = logical(new_DOUT_state); % Impose binary values
                L.set_dout;
            else
                error('Number of states must match number of ports in the list.')
            end
        end
        function set.scan_rate_requested_Hz(L,new_scan_rate_requested_Hz)
            % TODO: automatic maximum stream rate selection (instead of hardcoded limit 2500Hz)
            stream_rate_limit_Hz = 100000;%2500; % TODO: Include this as device dependent property...
            L.scan_rate_requested_Hz = new_scan_rate_requested_Hz;
            if L.stream_rate_requested_Hz > stream_rate_limit_Hz
                L.scan_rate_requested_Hz = floor(stream_rate_limit_Hz/L.number_of_AIN);
                fprintf(1,'Stream rate limited to %uHz, resulting scan rate is %uHz\n',L.stream_rate_requested_Hz,L.scan_rate_requested_Hz);
            end
        end
        
    end
    methods % GET-METHODS
        function duty_cycle = get.PWM_duty_cycle(L)
            % NOTE: The PWM_state is always between 0 and 1, where 0
            % corresponds to the lower duty cycle limit and 1 corresponds
            % to the upper limit. If no limits are specified, the duty
            % cycle limits are 0 and 1 (i.e. 0% and 100%).
            % For example, when controlling a servo, the duty cycle is
            % limited such that the pulse duration is between 1ms and 2ms.
            % At 50Hz carrier frequency, this implies duty cycle limits of
            % 0.05 and 0.10 (i.e. 5% and 10%). A PWM_state value of 0 then
            % corresponds with 1ms (minimum position) and 1 corresponds
            % with 2ms (maximum position).
            % E.g. the Parallax Continuous Rotation Servo (#900-00008) has
            % a range of 1.3 to 1.7ms, centered at 1.5ms, and preferred
            % carrier frequency of max. 50Hz ("... for smooth rotation, the
            % servo needs a 20ms pause between pulses."). Duty cycle limits
            % can be calculated as follows: 
            % For the T7: servo_pulse_limits_ms / PWM_carrier_period_ms
            % For the UE9: (1000-servo_pulse_limits_ms) / PWM_carrier_period_ms
            % This servo has linear response to duty cycle, with speed from
            % 0 to approx. +/-50rpm.
            if isempty(L.PWM_duty_cycle_limits)
                L.PWM_duty_cycle_limits = [0 1];
            end
            duty_cycle = L.PWM_duty_cycle_limits(1) + ...
                L.PWM_state * range(L.PWM_duty_cycle_limits);
        end
        function ast = get.AIN_state_V(L)
            if L.stream_started
                disp('Cannot poll: stream is running')
            else
                if ismember(L.device_type, {'UE9', 'U3'})
                    for i = 1:length(L.AIN_list)
                        [~, ast(i)] = L.UDobject.eGet(L.lj_handle,...
                            LabJack.LabJackUD.IO.GET_AIN, L.AIN_list(i), 0, 0);
                    end
                else % if strcmp(L.device_type, 'T7')
                    AIN_V = NET.createArray('System.Double', L.number_of_AIN); % This can be seen as a pointer... (LJM functions use pass by reference)
                    LabJack.LJM.eReadNames(L.lj_handle, L.number_of_AIN,...
                        L.AIN_names, AIN_V, 0);
                    ast = double(AIN_V);
                end
            end
        end
        function cn = get.AIN_names(L)
            cn = cell(1,L.number_of_AIN);
            for ch = 1:L.number_of_AIN
                cn{ch} = sprintf('AIN%u',L.AIN_list(ch));
            end
        end
        function noc = get.number_of_AIN(L)
            noc = length(L.AIN_list); % [-]
        end
        function nosc = get.number_of_stream_channels(L)
            nosc = L.number_of_AIN + ~isempty(L.DIN_list); % NOTE: all digital ports are combined into one stream channel
        end
        function srr = get.stream_rate_requested_Hz(L)
            srr = L.scan_rate_requested_Hz*L.number_of_AIN; % The actual number of samples that will be streamed
        end
        function escpc = get.expected_scans_per_chunk(L)
            escpc = L.scan_rate_requested_Hz * L.chunk_time_s;
        end
        function esapc = get.expected_samples_per_chunk(L)
            esapc = L.number_of_AIN*L.expected_scans_per_chunk;
        end
        function bs = get.buffer_size(L)
            bs = L.scan_rate_requested_Hz * L.buffer_time_s * L.number_of_AIN; % [-]
        end
        function dst = get.DIN_state(L)
            if L.stream_started
                disp('Cannot poll: stream is running')
            else
                switch L.device_type
                    case 'UE9'
                        for i = 1:length(L.DIN_list)
                            % This configures for input and gets the state
                            % (could also use GET_DIGITAL_BIT_STATE, which
                            % does not configure the port)
                            [~, dst(i)] = L.UDobject.eGet(L.lj_handle, ...
                                LabJack.LabJackUD.IO.GET_DIGITAL_BIT, ...
                                L.DIN_list(i), 0, 0); % TODO: Does this work? Also check which approach is most efficient: eGet or addrequest
                        end
                    case 'T7'
                        warning('get.DIN_state not implemented yet for T7')
                end
            end
        end
    end
    
    methods (Access = public, Hidden = false)
        
        % SHOW AIN
        show_AIN(L)
        
        % START STREAM
        start_stream(L)
        
        % STOP STREAM
        stop_stream(L)
        
        % PLOT STREAM (Q&D...)
        plot_stream(L,filterwindow)
        
        % DESTRUCTOR (NOTE: if a timer is used with handle to LJclass-object, delete needs to be called manually. This is not necessary of the timer is deleted beforehand (e.g. by the stop_stream() method) )
        function delete(L)
            if ~ischar(L.device_id) % If not labjack was found, device_id is still 'ANY'
                if L.is_running
                    L.stop_stream
                end
                delete(L.timer_handle)
                disp('L timer cleared')
                % TODO: Check timer deletion. Maybe it is safer to use delete(timerfindall)...
                
                if length(superclasses(L)) == 1 % NOTE: this is a Q&D solution to ensure that these calls are only made when L is not a subclass of LJclass (such as FestoClass)
                    L.DAC_state_V = zeros(size(L.DAC_list));
                    L.DOUT_state = zeros(size(L.DOUT_list));
                    L.relay_state = zeros(size(L.relay_list));
                    
                end
                L.PWM_list = [];
                
                if strcmp(L.device_type,'T7')
                    LabJack.LJM.Close(L.lj_handle); % NOTE: The UD.NET counterpart of the close function causes matlab to crash...
                    disp('Labjack closed')
                end
            end
        end
        
        
        
    end
    
    methods (Static, Access = private)
        
        % Callback for the timer (gathers data, updates plot during stream)
        timerfcn(~,~,L)
        
    end
    
    methods (Static, Access=public)
        
        % Moving average filter
        outputsignal = maf(inputsignal,filterwindow)
        
    end
    
    
    
    methods (Access = private, Hidden) % Use protected or private?
        
        % OPEN DEVICE
        open_labjack(L)
        
        % CONFIGURATION
        configure_labjack(L)
        
        % GET STREAM DATA
        get_stream_data(L)
        
        % TOGGLE RELAY
        toggle_relay(L)
        
        % SET PWM DUTY CYCLE
        update_PWM_duty_cycle(L)
        
        % UPDATE PWM CARRIER FREQUENCY
        actual_frequency_Hz = ...
            update_PWM_carrier_frequency(L, desired_carrier_frequency_Hz)
        
        % SET DIGITAL OUT
        set_dout(L)
        
        % SET ANALOG OUT
        set_aout(L)
        
        
        % Stream is running?
        function flag = is_running(L)
            % Try to get stream data (if not running, this yields an error
            % in L.last_error), to check whether the labjack hardware
            % stream is running. NOTE: this does not work while the matlab
            % timer is active because it blocks execution somehow.
            % Therefore, there is also a flag property called
            % "stream_started" to indicate whether the start_stream method
            % is active. Do not confuse the two: is_running is to be used
            % to check hardware streaming mode, whereas stream_started is
            % to be used to check if the software streaming function is
            % active.
            try
                L.get_stream_data()
                flag = 1;
            catch
                flag = 0;
            end
            %flag = strcmp(char(L.last_error),'NOERROR'); % If stream is running, get_stream_data should return 'NOERROR' (note that L.last_error is a labjack.labjackUD.LJUDERROR object, so need to cast as char)
        end % is running
        
        % SHOW ERROR MESSAGE
        function show_error_message(errmsg)
            if isa(errmsg,'NET.NetException')
                eNet = errmsg.ExceptionObject;
                if isa(eNet, 'LabJack.LabJackUD.LabJackUDException')
                    disp(['Error message:\n' eNet.LJUDError])
                end
            end
        end
        
        %         % ALLOWABLE STREAM RATE
        %         function max_stream_rate = labjack_maximum_stream_rate(resolution_bits)
        %             % According to Table 3-4, p.46 of the labjack ue9 users guide
        %             maximum_stream_rates = [50 16 4 1 0.25]*1000; % LabJack UE9 maximum stream rates [Hz]
        %             resolutions = 12:16; % LabJack UE9 resolution [bits] corresponding to max stream rates
        %             max_stream_rate = maximum_stream_rates(resolutions==resolution_bits);
        %
        %         end
        
    end % Private methods
    
    
    %% Backward compatibility
    
    properties (Hidden, Dependent)
        % Properties that have been renamed
        FIO_list % Renamed to DOUT_list
        FIO_state % Renamed to DOUT_state
    end
    properties (Hidden)
        % Properties that have been removed
    end
    methods
        % These methods are usually only invoked e.g. when old object is
        % loaded or old code is used.
        function set.FIO_list(L,value)
            warning('legacy set FIO_list')
            L.DOUT_list = value;
        end
        function value = get.FIO_list(E)
            warning('legacy get FIO_list')
            value = E.DOUT_list;
        end
        function set.FIO_state(L,value)
            warning('legacy set FIO_state')
            L.DOUT_state = value;
        end
        function value = get.FIO_state(E)
            warning('legacy get FIO_state')
            value = E.DOUT_state;
        end
    end
    
    
end % Classdef
