function open_labjack(L)

% NOTE:
%   The AddRequest and GetResult functions have arguments:
%       Handle  This is an input to all request/result functions that tells the function what LabJack it is talking to. The handle is obtained from the OpenLabJack function.
%       IOType  This is an input to all request/result functions that specifies what type of action is being done.
%       Channel  This is an input to all request/result functions that generally specifies which channel of I/O is being written/read, although with the config IOTypes special constants are passed for channel to specify what is being configured.
%       Value  This is an input or output to all request/result functions that is used to write or read the value for the item being operated on.
%       x1  This parameter is only used in some of the request/result functions, and is used when extra information is needed for certain IOTypes.
%       UserData  This parameter is only used in some of the request/result functions, and is data that is simply passed along with the request, and returned unmodified by the result. Can be used to store any sort of information with the request, to allow a generic parser to determine what should be done when the results are received.
%
%   The ePut and eGet functions have arguments:
%       Handle  Handle returned by OpenLabJack().
%       IOType  The type of request. See Section 4.3.
%       Channel  The channel number of the particular IOType.
%       pValue  Pointer to Value sends and receives data.
%       x1  Optional parameter used by some IOTypes.
%
% NOTE: The onboard UE9 buffer (4 Mbit) can hold over 180,000 samples.
% NOTE: A sample is defined as a single conversion of a single channel,
%       while a scan is defined as a single conversion of all channels
%       being acquired. That means that according to Table 3-4 the maximum
%       scan rate for a stream of five 12-bit channels is 50k/5 = 10 k/second.
% NOTE: Buffer overrun while streaming: That means the driver is transferring
%       the data from the UE9 (Comm buffer) to the PC (UD buffer) fast enough,
%       but you are not getting the data out of the UD buffer fast enough.
%
% Author: DJ van Gerwen
% Created: 2016 june 5

if isempty(L.device_type)
    % Get both assemblies
    L = getAssembly(L, 'all');
    
    % Check for any connected Labjack devices
    device_found = zeros(size(L.supported_device_types));
    for i = 1:length(L.supported_device_types)
        if strcmp(L.supported_device_types{i}, 'T7')
            % Check for T7
            [L.last_error, device_found(i)] = ...
                LabJack.LJM.ListAllS(L.supported_device_types{i}, ...
                L.connection_type, 0, 0, 0, 0, 0);
        else
            % Check for UE9 or U3
            [L.last_error, device_found(i)] = L.UDobject.ListAll( ...
                LabJack.LabJackUD.DEVICE.(L.supported_device_types{i}), ...
                LabJack.LabJackUD.CONNECTION.(L.connection_type), 0, 0, 0, 0);
        end
    end
    
    % Assign device type (assumes no more than one connected device)
    L.device_type = L.supported_device_types{logical(device_found)};
    
end

switch L.device_type
    
    case 'T7'
        
        % Get assembly
        if isempty(L.Massembly)
            L = getAssembly(L, 'LJM');
        end
        
        % Open labjack T7 device
        [L.last_error, L.lj_handle] = ...
            LabJack.LJM.OpenS(L.device_type,L.connection_type,L.device_id, 0);
        
        % Get device specific information (output arguments are: error, device type, connection type, serial number, IP address, port, maxbytesperMB)
        [L.last_error, ~, ~, L.device_id, ~, ~, ~] = ...
            LabJack.LJM.GetHandleInfo(L.lj_handle, 0, 0, 0, 0, 0, 0);
        
    otherwise % U3 or UE9
        
        % Get assembly
        if isempty(L.UDassembly)
            L = getAssembly(L, 'UDdotNET');
        end
        
        % Open LabJack U3 or UE9 device using specified connection.
        [L.last_error, L.lj_handle] = ...
            L.UDobject.OpenLabJack(LabJack.LabJackUD.DEVICE.(L.device_type), ...
            LabJack.LabJackUD.CONNECTION.(L.connection_type), '0', true, 0);
end

end

function L = getAssembly(L, type)

if ismember(type,{'LJM', 'all'})
    % Make the LJM assembly visible in MATLAB
    L.Massembly = NET.addAssembly('LabJack.LJM'); %Make the LJM .NET assembly visible in MATLAB
    L.Mconstants =  System.Activator.CreateInstance( ...
        L.Massembly.AssemblyHandle.GetType('LabJack.LJM+CONSTANTS')); %creating an object to nested class LabJack.LJM.CONSTANTS
end
if ismember(type,{'UDdotNET', 'all'})
    % Make the UD .NET assembly visible in MATLAB and instantiate an object
    L.UDassembly = NET.addAssembly('LJUDDotNet');
    L.UDobject = LabJack.LabJackUD.LJUD;
end

end