function actual_carrier_frequency_Hz = update_PWM_carrier_frequency(L, desired_carrier_frequency_Hz)
% Update PWM carrier frequency
%
% NOTE: PWM_carrier_frequency = clock_frequency/divisor/roll_value,
% so roll_value = clock_frequency/PWM_frequency/divisor
%
% Author: DJ van Gerwen
% Created: 2016 june 5


switch L.device_type
    case 'T7'
        
        % Labjack T7 constants
        % The T7 clock resets when it reaches the roll value (maximum is
        % 2^32-1 for 32-bit clock). The roll value determines the
        % resolution, so should be maximized. Thus, use minimum divisor
        % with maximum roll value.
        L.PWM_clock_source = 0; % Labjack T7 clock0 is 32bit, so counts from 0 to roll_value (maximum 2^32-1) at 80MHz
        L.PWM_clock_frequency_Hz = 80e6; % Labjack T7 clock0 frequency is 80MHz
        L.PWM_clock_divisor = 1; % A divisor of 1 allows frequencies from 80e6/(2^32-1)=0.019Hz to 80MHz
       
        if isempty(L.PWM_list) % NOT sure if disabling is necessary here...
            LabJack.LJM.eWriteName(L.lj_handle, ...
                sprintf('DIO_EF_CLOCK%u_ENABLE',L.PWM_clock_source), 0); % Disable clock source
        end
        L.PWM_clock_roll_value = min(2^32-1, ... % and impose maximum value of 2^32-1 for 32-bit clock
            floor(L.PWM_clock_frequency_Hz / L.PWM_clock_divisor / ...
            desired_carrier_frequency_Hz));
        
        % Update the carrier frequency to match the "achieved" frequency.
        actual_carrier_frequency_Hz = L.PWM_clock_frequency_Hz / ...
            L.PWM_clock_divisor / L.PWM_clock_roll_value;
        
        % Write to labjack device
        LabJack.LJM.eWriteName(L.lj_handle, ...
            sprintf('DIO_EF_CLOCK%u_DIVISOR',L.PWM_clock_source),...
            L.PWM_clock_divisor); % Set divisor
        LabJack.LJM.eWriteName(L.lj_handle, ...
            sprintf('DIO_EF_CLOCK%u_ROLL_VALUE',L.PWM_clock_source),...
            L.PWM_clock_roll_value); % Set roll value
        if isempty(L.PWM_list) % NOT sure if enabling is necessary here...
            LabJack.LJM.eWriteName(L.lj_handle, ...
                sprintf('DIO_EF_CLOCK%u_ENABLE',L.PWM_clock_source), 1); % Enable clock source
        end
        
    case 'UE9'
        
        % Labjack UE9 Constants
        LJ_tmPWM16 = 0; % Timer mode 0 gives 16 bit resolution
        timer_resolution_bits = 16;
        UE9_source_clocks_Hz = [750e3 48e6];
        L.PWM_clock_source = 1; 
        L.PWM_clock_frequency_Hz = UE9_source_clocks_Hz(2); % 48MHz allows carrier freq. from 3-730 Hz)
        L.PWM_clock_roll_value = 2^timer_resolution_bits - 1; % ( "roll value" is T7 terminology...)
        
        % The clock divisor depends on the desired carrier frequency
        desired_divisor = round( L.PWM_clock_frequency_Hz / ...
            L.PWM_clock_roll_value / desired_carrier_frequency_Hz);
        if desired_divisor > 255
            desired_divisor = 0; % 0 actually divides by 256
        end
        % Allowable range is 0 to 255 (0 divides by 256)
        L.PWM_clock_divisor = max(desired_divisor, 0); % Labjack frequency divisor 
        
        % Update the carrier frequency to match the "achieved" frequency.
        actual_carrier_frequency_Hz = L.PWM_clock_frequency_Hz / ...
            L.PWM_clock_divisor / 2^timer_resolution_bits;
                
        % Set 48MHz Clock  (this holds for all timers) 
        L.UDobject.AddRequest(L.lj_handle, LabJack.LabJackUD.IO.PUT_CONFIG, ...
            LabJack.LabJackUD.CHANNEL.TIMER_CLOCK_BASE, L.PWM_clock_source, 0, 0);
        
        % Carrier frequency = 48MHz / 18(i.e. divisor) / 2^16(i.e. mode 0) =
        % 40.69Hz (this holds for all timers)
        L.UDobject.AddRequest(L.lj_handle, LabJack.LabJackUD.IO.PUT_CONFIG, ...
            LabJack.LabJackUD.CHANNEL.TIMER_CLOCK_DIVISOR, L.PWM_clock_divisor, 0, 0);
        
        % Enable Timers on requested FIO ports
        L.UDobject.AddRequest(L.lj_handle, LabJack.LabJackUD.IO.PUT_CONFIG, ...
            LabJack.LabJackUD.CHANNEL.NUMBER_TIMERS_ENABLED, ...
            length(L.PWM_list), 0, 0);
        
        % Set Timer Mode to 16-bit PWM (65536 steps per carrier period) for
        % all designated FIO ports
        for FIOi = L.PWM_list(:)' % NOTE: (:)' enforces a row vector
            L.UDobject.AddRequest(L.lj_handle, LabJack.LabJackUD.IO.PUT_TIMER_MODE, ...
                FIOi, LJ_tmPWM16, 0, 0);
        end
        % TODO: Maybe move this part to the set.PWM_list method?
        
        % Execute above requests
        L.UDobject.GoOne(L.lj_handle); 
        
        % Run while loop until no-more-data-available-error is
        % returned, to ensure that the device has fully configured its
        % channels before continuing.
        [~] = L.UDobject.GetFirstResult(L.lj_handle, LabJack.LabJackUD.IO, ...
            LabJack.LabJackUD.CHANNEL, 0, 0, 0);
        finished = 0;
        while ~finished
            try
                [~] = L.UDobject.GetNextResult(L.lj_handle, LabJack.LabJackUD.IO, ...
                    LabJack.LabJackUD.CHANNEL, 0, 0, 0);
            catch e
                if isa(e, 'NET.NetException')
                    eNet = e.ExceptionObject;
                    if isa(eNet, 'LabJack.LabJackUD.LabJackUDException')
                        if(eNet.LJUDError == ...
                                LabJack.LabJackUD.LJUDERROR.NO_MORE_DATA_AVAILABLE)
                            finished = 1;
                        end
                    end
                end
                %Report non NO_MORE_DATA_AVAILABLE error.
                if ~finished
                    throw(e)
                end
            end
        end
    otherwise
        error('PWM not implemented yet for U3')
end

end